@@ -12,18 +12,42 @@
 # See the License for the specific language governing permissions and
 # limitations under the License.
 
+"""All of Nox's configuration options."""
+
 from __future__ import annotations
 
 import argparse
 import functools
+import itertools
 import os
 import sys
-from typing import Any, Sequence
+from typing import TYPE_CHECKING, Any, Callable, Literal, Sequence
+
+import argcomplete
 
 from nox import _option_set
 from nox.tasks import discover_manifest, filter_manifest, load_nox_module
+from nox.virtualenv import ALL_VENVS
 
-"""All of Nox's configuration options."""
+if TYPE_CHECKING:
+    from collections.abc import Iterable
+
+    from nox._option_set import NoxOptions
+
+
+__all__ = ["ReuseVenvType", "noxfile_options", "options"]
+
+
+def __dir__() -> list[str]:
+    return __all__
+
+
+# User-specified arguments will be a regular string
+class DefaultStr(str):
+    __slots__ = ()
+
+
+ReuseVenvType = Literal["no", "yes", "never", "always"]
 
 options = _option_set.OptionSet(
     description="Nox is a Python automation toolkit.", add_help=False
@@ -64,11 +88,11 @@ options.add_groups(
 )
 
 
-def _sessions_and_keywords_merge_func(
-    key: str, command_args: argparse.Namespace, noxfile_args: argparse.Namespace
+def _sessions_merge_func(
+    key: str, command_args: argparse.Namespace, noxfile_args: NoxOptions
 ) -> list[str]:
-    """Only return the Noxfile value for sessions/keywords if neither sessions
-    or keywords are specified on the command-line.
+    """Only return the Noxfile value for sessions/keywords if neither sessions,
+    keywords or tags are specified on the command-line.
 
     Args:
         key (str): This function is used for both the "sessions" and "keywords"
@@ -76,22 +100,26 @@ def _sessions_and_keywords_merge_func(
             same function for both options.
         command_args (_option_set.Namespace): The options specified on the
             command-line.
-        noxfile_Args (_option_set.Namespace): The options specified in the
+        noxfile_Args (NoxOptions): The options specified in the
             Noxfile."""
-    if not command_args.sessions and not command_args.keywords:
+    if (
+        not command_args.sessions
+        and not command_args.keywords
+        and not command_args.tags
+    ):
         return getattr(noxfile_args, key)  # type: ignore[no-any-return]
     return getattr(command_args, key)  # type: ignore[no-any-return]
 
 
 def _default_venv_backend_merge_func(
-    command_args: argparse.Namespace, noxfile_args: argparse.Namespace
+    command_args: argparse.Namespace, noxfile_args: NoxOptions
 ) -> str:
     """Merge default_venv_backend from command args and Noxfile. Default is "virtualenv".
 
     Args:
         command_args (_option_set.Namespace): The options specified on the
             command-line.
-        noxfile_Args (_option_set.Namespace): The options specified in the
+        noxfile_Args (NoxOptions): The options specified in the
             Noxfile.
     """
     return (
@@ -102,14 +130,14 @@ def _default_venv_backend_merge_func(
 
 
 def _force_venv_backend_merge_func(
-    command_args: argparse.Namespace, noxfile_args: argparse.Namespace
+    command_args: argparse.Namespace, noxfile_args: NoxOptions
 ) -> str:
     """Merge force_venv_backend from command args and Noxfile. Default is None.
 
     Args:
         command_args (_option_set.Namespace): The options specified on the
             command-line.
-        noxfile_Args (_option_set.Namespace): The options specified in the
+        noxfile_Args (NoxOptions): The options specified in the
             Noxfile.
     """
     if command_args.no_venv:
@@ -117,37 +145,68 @@ def _force_venv_backend_merge_func(
             command_args.force_venv_backend is not None
             and command_args.force_venv_backend != "none"
         ):
-            raise ValueError(
-                "You can not use `--no-venv` with a non-none `--force-venv-backend`"
-            )
-        else:
-            return "none"
-    else:
-        return command_args.force_venv_backend or noxfile_args.force_venv_backend  # type: ignore[no-any-return]
+            msg = "You can not use `--no-venv` with a non-none `--force-venv-backend`"
+            raise ValueError(msg)
+        return "none"
+    return command_args.force_venv_backend or noxfile_args.force_venv_backend  # type: ignore[return-value]
 
 
 def _envdir_merge_func(
-    command_args: argparse.Namespace, noxfile_args: argparse.Namespace
+    command_args: argparse.Namespace, noxfile_args: NoxOptions
 ) -> str:
     """Ensure that there is always some envdir.
 
     Args:
         command_args (_option_set.Namespace): The options specified on the
             command-line.
-        noxfile_Args (_option_set.Namespace): The options specified in the
+        noxfile_Args (NoxOptions): The options specified in the
             Noxfile.
     """
-    return command_args.envdir or noxfile_args.envdir or ".nox"
+    return os.fspath(command_args.envdir or noxfile_args.envdir or ".nox")
 
 
-def _sessions_default() -> list[str] | None:
-    """Looks at the NOXSESSION env var to set the default value for sessions."""
-    nox_env = os.environ.get("NOXSESSION")
-    env_sessions = nox_env.split(",") if nox_env else None
-    return env_sessions
+def _reuse_venv_merge_func(
+    command_args: argparse.Namespace, noxfile_args: NoxOptions
+) -> ReuseVenvType:
+    """Merge reuse_venv from command args and Noxfile while maintaining
+    backwards compatibility with reuse_existing_virtualenvs. Default is "no".
+
+    Args:
+        command_args (_option_set.Namespace): The options specified on the
+            command-line.
+        noxfile_Args (NoxOptions): The options specified in the
+            Noxfile.
+    """
+    # back-compat scenario with no_reuse_existing_virtualenvs/reuse_existing_virtualenvs
+    if command_args.no_reuse_existing_virtualenvs:
+        return "no"
+    if (
+        command_args.reuse_existing_virtualenvs
+        or noxfile_args.reuse_existing_virtualenvs
+    ):
+        return "yes"
+    # regular option behavior
+    return command_args.reuse_venv or noxfile_args.reuse_venv or "no"
 
 
-def _color_finalizer(value: bool, args: argparse.Namespace) -> bool:
+def default_env_var_list_factory(env_var: str) -> Callable[[], list[str] | None]:
+    """Looks at the env var to set the default value for a list of env vars.
+
+    Args:
+        env_var (str): The name of the environment variable to look up.
+
+    Returns:
+        A callback that retrieves a list from a comma-delimited environment variable.
+    """
+
+    def _default_list() -> list[str] | None:
+        env_value = os.environ.get(env_var)
+        return env_value.split(",") if env_value else None
+
+    return _default_list
+
+
+def _color_finalizer(_value: bool, args: argparse.Namespace) -> bool:  # noqa: FBT001
     """Figures out the correct value for "color" based on the two color flags.
 
     Args:
@@ -157,15 +216,15 @@ def _color_finalizer(value: bool, args: argparse.Namespace) -> bool:
     Returns:
         The new value for the "color" option.
     """
-    if args.forcecolor is True and args.nocolor is True:
+    if args.forcecolor and args.nocolor:
         raise _option_set.ArgumentError(
             None, "Can not specify both --no-color and --force-color."
         )
 
-    if args.forcecolor is True:
+    if args.forcecolor:
         return True
 
-    if args.nocolor is True:
+    if args.nocolor or "NO_COLOR" in os.environ:
         return False
 
     return sys.stdout.isatty()
@@ -180,15 +239,26 @@ def _force_pythons_finalizer(
     return value
 
 
-def _R_finalizer(value: bool, args: argparse.Namespace) -> bool:
-    """Propagate -R to --reuse-existing-virtualenvs and --no-install."""
+def _R_finalizer(value: bool, args: argparse.Namespace) -> bool:  # noqa: FBT001
+    """Propagate -R to --reuse-existing-virtualenvs and --no-install and --reuse-venv=yes."""
     if value:
+        args.reuse_venv = "yes"
         args.reuse_existing_virtualenvs = args.no_install = value
     return value
 
 
+def _reuse_existing_virtualenvs_finalizer(
+    value: bool,  # noqa: FBT001
+    args: argparse.Namespace,
+) -> bool:
+    """Propagate --reuse-existing-virtualenvs to --reuse-venv=yes."""
+    if value:
+        args.reuse_venv = "yes"
+    return value
+
+
 def _posargs_finalizer(
-    value: Sequence[Any], args: argparse.Namespace
+    value: Sequence[Any], _args: argparse.Namespace
 ) -> Sequence[Any] | list[Any]:
     """Removes the leading "--"s in the posargs array (if any) and asserts that
     remaining arguments came after a "--".
@@ -213,18 +283,48 @@ def _posargs_finalizer(
     return posargs[dash_index + 1 :]
 
 
+def _python_completer(
+    prefix: str,  # noqa: ARG001
+    parsed_args: argparse.Namespace,
+    **kwargs: Any,
+) -> Iterable[str]:
+    module = load_nox_module(parsed_args)
+    manifest = discover_manifest(module, parsed_args)
+    return filter(
+        None,
+        (
+            session.func.python  # type:ignore[misc] # str sequences flattened, other non-strs falsey and filtered out
+            for session, _ in manifest.list_all_sessions()
+        ),
+    )
+
+
 def _session_completer(
-    prefix: str, parsed_args: argparse.Namespace, **kwargs: Any
-) -> list[str]:
-    global_config = parsed_args
-    module = load_nox_module(global_config)
-    manifest = discover_manifest(module, global_config)
-    filtered_manifest = filter_manifest(manifest, global_config)
+    prefix: str,  # noqa: ARG001
+    parsed_args: argparse.Namespace,
+    **kwargs: Any,
+) -> Iterable[str]:
+    parsed_args.list_sessions = True
+    module = load_nox_module(parsed_args)
+    manifest = discover_manifest(module, parsed_args)
+    filtered_manifest = filter_manifest(manifest, parsed_args)
     if isinstance(filtered_manifest, int):
         return []
-    return [
+    return (
         session.friendly_name for session, _ in filtered_manifest.list_all_sessions()
-    ]
+    )
+
+
+def _tag_completer(
+    prefix: str,  # noqa: ARG001
+    parsed_args: argparse.Namespace,
+    **kwargs: Any,
+) -> Iterable[str]:
+    module = load_nox_module(parsed_args)
+    manifest = discover_manifest(module, parsed_args)
+    return itertools.chain.from_iterable(
+        filter(None, (session.tags for session, _ in manifest.list_all_sessions()))
+    )
 
 
 options.add_options(
@@ -243,6 +343,18 @@ options.add_options(
         action="store_true",
         help="Show the Nox version and exit.",
     ),
+    _option_set.Option(
+        "script_mode",
+        "--script-mode",
+        group=options.groups["general"],
+        choices=["none", "fresh", "reuse"],
+        default="reuse",
+    ),
+    _option_set.Option(
+        "script_venv_backend",
+        "--script-venv-backend",
+        group=options.groups["general"],
+    ),
     _option_set.Option(
         "list_sessions",
         "-l",
@@ -252,6 +364,13 @@ options.add_options(
         action="store_true",
         help="List all available sessions and exit.",
     ),
+    _option_set.Option(
+        "json",
+        "--json",
+        group=options.groups["sessions"],
+        action="store_true",
+        help="JSON output formatting. Requires list-sessions currently.",
+    ),
     _option_set.Option(
         "sessions",
         "-s",
@@ -260,9 +379,9 @@ options.add_options(
         "--session",
         group=options.groups["sessions"],
         noxfile=True,
-        merge_func=functools.partial(_sessions_and_keywords_merge_func, "sessions"),
+        merge_func=functools.partial(_sessions_merge_func, "sessions"),
         nargs="*",
-        default=_sessions_default,
+        default=default_env_var_list_factory("NOXSESSION"),
         help="Which sessions to run. By default, all sessions will run.",
         completer=_session_completer,
     ),
@@ -274,7 +393,9 @@ options.add_options(
         group=options.groups["python"],
         noxfile=True,
         nargs="*",
+        default=default_env_var_list_factory("NOXPYTHON"),
         help="Only run sessions that use the given python interpreter versions.",
+        completer=_python_completer,
     ),
     _option_set.Option(
         "keywords",
@@ -282,8 +403,9 @@ options.add_options(
         "--keywords",
         group=options.groups["sessions"],
         noxfile=True,
-        merge_func=functools.partial(_sessions_and_keywords_merge_func, "keywords"),
+        merge_func=functools.partial(_sessions_merge_func, "keywords"),
         help="Only run sessions that match the given expression.",
+        completer=argcomplete.completers.ChoicesCompleter(()),  # type: ignore[no-untyped-call]
     ),
     _option_set.Option(
         "tags",
@@ -291,8 +413,10 @@ options.add_options(
         "--tags",
         group=options.groups["sessions"],
         noxfile=True,
+        merge_func=functools.partial(_sessions_merge_func, "tags"),
         nargs="*",
         help="Only run sessions with the given tags.",
+        completer=_tag_completer,
     ),
     _option_set.Option(
         "posargs",
@@ -302,14 +426,12 @@ options.add_options(
         help="Arguments following ``--`` that are passed through to the session(s).",
         finalizer_func=_posargs_finalizer,
     ),
-    _option_set.Option(
+    *_option_set.make_flag_pair(
         "verbose",
-        "-v",
-        "--verbose",
+        ("-v", "--verbose"),
+        ("--no-verbose",),
         group=options.groups["reporting"],
-        action="store_true",
         help="Logs the output of all commands run including commands marked silent.",
-        noxfile=True,
     ),
     _option_set.Option(
         "add_timestamp",
@@ -318,7 +440,6 @@ options.add_options(
         group=options.groups["reporting"],
         action="store_true",
         help="Adds a timestamp to logged output.",
-        noxfile=True,
     ),
     _option_set.Option(
         "default_venv_backend",
@@ -326,13 +447,13 @@ options.add_options(
         "--default-venv-backend",
         group=options.groups["environment"],
         noxfile=True,
+        default=lambda: os.environ.get("NOX_DEFAULT_VENV_BACKEND"),
         merge_func=_default_venv_backend_merge_func,
         help=(
             "Virtual environment backend to use by default for Nox sessions, this is"
-            " ``'virtualenv'`` by default but any of ``('virtualenv', 'conda', 'mamba',"
-            " 'venv')`` are accepted."
+            f" ``'virtualenv'`` by default but any of ``{list(ALL_VENVS)!r}`` are accepted."
         ),
-        choices=["none", "virtualenv", "conda", "mamba", "venv"],
+        choices=list(ALL_VENVS),
     ),
     _option_set.Option(
         "force_venv_backend",
@@ -344,10 +465,9 @@ options.add_options(
         help=(
             "Virtual environment backend to force-use for all Nox sessions in this run,"
             " overriding any other venv backend declared in the Noxfile and ignoring"
-            " the default backend. Any of ``('virtualenv', 'conda', 'mamba', 'venv')``"
-            " are accepted."
+            f" the default backend. Any of ``{list(ALL_VENVS)!r}`` are accepted."
         ),
-        choices=["none", "virtualenv", "conda", "mamba", "venv"],
+        choices=list(ALL_VENVS),
     ),
     _option_set.Option(
         "no_venv",
@@ -360,6 +480,18 @@ options.add_options(
             " creating a venv. This is an alias for '--force-venv-backend none'."
         ),
     ),
+    _option_set.Option(
+        "reuse_venv",
+        "--reuse-venv",
+        group=options.groups["environment"],
+        noxfile=True,
+        merge_func=_reuse_venv_merge_func,
+        help=(
+            "Controls existing virtualenvs recreation. This is ``'no'`` by"
+            " default, but any of ``('yes', 'no', 'always', 'never')`` are accepted."
+        ),
+        choices=["yes", "no", "always", "never"],
+    ),
     *_option_set.make_flag_pair(
         "reuse_existing_virtualenvs",
         ("-r", "--reuse-existing-virtualenvs"),
@@ -368,7 +500,8 @@ options.add_options(
             "--no-reuse-existing-virtualenvs",
         ),
         group=options.groups["environment"],
-        help="Re-use existing virtualenvs instead of recreating them.",
+        help="This is an alias for '--reuse-venv=yes|no'.",
+        finalizer_func=_reuse_existing_virtualenvs_finalizer,
     ),
     _option_set.Option(
         "R",
@@ -377,7 +510,7 @@ options.add_options(
         group=options.groups["environment"],
         action="store_true",
         help=(
-            "Re-use existing virtualenvs and skip package re-installation."
+            "Reuse existing virtualenvs and skip package re-installation."
             " This is an alias for '--reuse-existing-virtualenvs --no-install'."
         ),
         finalizer_func=_R_finalizer,
@@ -387,7 +520,7 @@ options.add_options(
         "-f",
         "--noxfile",
         group=options.groups["general"],
-        default="noxfile.py",
+        default=DefaultStr("noxfile.py"),
         help="Location of the Python file containing Nox sessions.",
     ),
     _option_set.Option(
@@ -397,6 +530,21 @@ options.add_options(
         merge_func=_envdir_merge_func,
         group=options.groups["environment"],
         help="Directory where Nox will store virtualenvs, this is ``.nox`` by default.",
+        completer=argcomplete.completers.DirectoriesCompleter(),  # type: ignore[no-untyped-call]
+    ),
+    _option_set.Option(
+        "download_python",
+        "--download-python",
+        "--download-python",
+        noxfile=True,
+        group=options.groups["python"],
+        default=lambda: os.getenv("NOX_DOWNLOAD_PYTHON"),
+        help=(
+            "When should nox download python standalone builds to run the sessions,"
+            " defaults to 'auto' which will download when the version requested can't"
+            " be found in the running environment."
+        ),
+        choices=["auto", "never", "always"],
     ),
     _option_set.Option(
         "extra_pythons",
@@ -404,19 +552,25 @@ options.add_options(
         "--extra-python",
         group=options.groups["python"],
         nargs="*",
+        default=default_env_var_list_factory("NOXEXTRAPYTHON"),
         help="Additionally, run sessions using the given python interpreter versions.",
+        completer=_python_completer,
     ),
     _option_set.Option(
         "force_pythons",
+        "-P",
         "--force-pythons",
         "--force-python",
         group=options.groups["python"],
         nargs="*",
+        default=default_env_var_list_factory("NOXFORCEPYTHON"),
         help=(
             "Run sessions with the given interpreters instead of those listed in the"
             " Noxfile. This is a shorthand for ``--python=X.Y --extra-python=X.Y``."
+            " It will also work on sessions that don't have any interpreter parametrized."
         ),
         finalizer_func=_force_pythons_finalizer,
+        completer=_python_completer,
     ),
     *_option_set.make_flag_pair(
         "stop_on_first_error",
@@ -431,7 +585,7 @@ options.add_options(
         ("--no-error-on-missing-interpreters",),
         group=options.groups["execution"],
         help="Error instead of skipping sessions if an interpreter can not be located.",
-        default="CI" in os.environ,
+        default=lambda: "CI" in os.environ,
     ),
     *_option_set.make_flag_pair(
         "error_on_external_run",
@@ -458,7 +612,7 @@ options.add_options(
         action="store_true",
         help=(
             "Skip invocations of session methods for installing packages"
-            " (session.install, session.conda_install, session.run_always)"
+            " (session.install, session.conda_install, session.run_install)"
             " when a virtualenv is being reused."
         ),
     ),
@@ -468,6 +622,7 @@ options.add_options(
         group=options.groups["reporting"],
         noxfile=True,
         help="Output a report of all sessions to the given filename.",
+        completer=argcomplete.completers.FilesCompleter(("json",)),  # type: ignore[no-untyped-call]
     ),
     _option_set.Option(
         "non_interactive",
@@ -510,7 +665,7 @@ options.add_options(
         "invoked_from",
         group=None,
         hidden=True,
-        default=lambda: os.getcwd(),
+        default=os.getcwd,
     ),
 )
 
