@@ -14,8 +14,8 @@
 
 from __future__ import annotations
 
-import argparse
 import contextlib
+import datetime
 import enum
 import hashlib
 import os
@@ -24,34 +24,51 @@ import re
 import shutil
 import subprocess
 import sys
+import time
 import unicodedata
-from collections.abc import (
-    Callable,
-    Generator,
-    Iterable,
-    Mapping,
-    Sequence,
-)
-from types import TracebackType
 from typing import (
     TYPE_CHECKING,
     Any,
     NoReturn,
 )
 
+import humanize
+
 import nox.command
 import nox.virtualenv
-from nox._decorators import Func
 from nox.logger import logger
 from nox.popen import DEFAULT_INTERRUPT_TIMEOUT, DEFAULT_TERMINATE_TIMEOUT
-from nox.virtualenv import CondaEnv, PassthroughEnv, ProcessEnv, VirtualEnv
+from nox.virtualenv import (
+    CondaEnv,
+    PassthroughEnv,
+    ProcessEnv,
+    VirtualEnv,
+    get_virtualenv,
+)
 
 if TYPE_CHECKING:
+    import argparse
+    from collections.abc import (
+        Callable,
+        Generator,
+        Iterable,
+        Iterator,
+        Mapping,
+        Sequence,
+    )
+    from types import TracebackType
     from typing import IO
 
+    from nox._decorators import Func
     from nox.command import ExternalType
     from nox.manifest import Manifest
 
+__all__ = ["Result", "Session", "SessionRunner", "Status", "nox"]
+
+
+def __dir__() -> list[str]:
+    return __all__
+
 
 @contextlib.contextmanager
 def _chdir(path: str) -> Generator[None, None, None]:
@@ -81,7 +98,7 @@ def _normalize_path(envdir: str, path: str | bytes) -> str:
     full_path = os.path.join(envdir, path)
     if len(full_path) > 100 - len("bin/pythonX.Y"):
         if len(envdir) < 100 - 9:
-            path = hashlib.sha1(path.encode("ascii")).hexdigest()[:8]
+            path = hashlib.sha1(path.encode("ascii")).hexdigest()[:8]  # noqa: S324
             full_path = os.path.join(envdir, path)
             logger.warning("The virtualenv name was hashed to avoid being too long.")
         else:
@@ -101,22 +118,20 @@ def _dblquote_pkg_install_args(args: Iterable[str]) -> tuple[str, ...]:
     def _dblquote_pkg_install_arg(pkg_req_str: str) -> str:
         # sanity check: we need an even number of double-quotes
         if pkg_req_str.count('"') % 2 != 0:
-            raise ValueError(
-                f"ill-formatted argument with odd number of quotes: {pkg_req_str}"
-            )
+            msg = f"ill-formatted argument with odd number of quotes: {pkg_req_str}"
+            raise ValueError(msg)
 
         if "<" in pkg_req_str or ">" in pkg_req_str:
             if pkg_req_str[0] == pkg_req_str[-1] == '"':
                 # already double-quoted string
                 return pkg_req_str
-            else:
-                # need to double-quote string
-                if '"' in pkg_req_str:
-                    raise ValueError(f"Cannot escape requirement string: {pkg_req_str}")
-                return f'"{pkg_req_str}"'
-        else:
-            # no dangerous char: no need to double-quote string
-            return pkg_req_str
+            # need to double-quote string
+            if '"' in pkg_req_str:
+                msg = f"Cannot escape requirement string: {pkg_req_str}"
+                raise ValueError(msg)
+            return f'"{pkg_req_str}"'
+        # no dangerous char: no need to double-quote string
+        return pkg_req_str
 
     # double-quote all args that need to be and return the result
     return tuple(_dblquote_pkg_install_arg(a) for a in args)
@@ -142,7 +157,7 @@ class _WorkingDirContext:
         self._prev_working_dir = os.getcwd()
         os.chdir(dir)
 
-    def __enter__(self) -> _WorkingDirContext:
+    def __enter__(self) -> _WorkingDirContext:  # noqa: PYI034
         return self
 
     def __exit__(
@@ -182,10 +197,15 @@ class Session:
         return self._runner.friendly_name
 
     @property
-    def env(self) -> dict[str, str]:
+    def env(self) -> dict[str, str | None]:
         """A dictionary of environment variables to pass into all commands."""
         return self.virtualenv.env
 
+    @property
+    def env_dir(self) -> pathlib.Path:
+        """The path to the environment of this session."""
+        return pathlib.Path(self._runner.envdir)
+
     @property
     def posargs(self) -> list[str]:
         """Any extra arguments from the ``nox`` commandline or :class:`Session.notify`."""
@@ -196,7 +216,8 @@ class Session:
         """The virtualenv that all commands are run in."""
         venv = self._runner.venv
         if venv is None:
-            raise ValueError("A virtualenv has not been created for this session")
+            msg = "A virtualenv has not been created for this session"
+            raise ValueError(msg)
         return venv
 
     @property
@@ -222,7 +243,8 @@ class Session:
         """The first bin directory for the virtualenv."""
         paths = self.bin_paths
         if paths is None:
-            raise ValueError("The environment does not have a bin directory.")
+            msg = "The environment does not have a bin directory."
+            raise ValueError(msg)
         return paths[0]
 
     def create_tmp(self) -> str:
@@ -273,6 +295,7 @@ class Session:
             silent=silent,
             success_codes=success_codes,
             external=None,
+            log=log,
             stdout=stdout,
             stderr=stderr,
             interrupt_timeout=interrupt_timeout,
@@ -434,7 +457,8 @@ class Session:
         :type stderr: file or file descriptor
         """
         if not args:
-            raise ValueError("At least one argument required to run().")
+            msg = "At least one argument required to run()."
+            raise ValueError(msg)
 
         if len(args) == 1 and isinstance(args[0], (list, tuple)):
             msg = "First argument to `session.run` is a list. Did you mean to use `session.run(*args)`?"
@@ -524,7 +548,8 @@ class Session:
             return None
 
         if not args:
-            raise ValueError("At least one argument required to run_install().")
+            msg = "At least one argument required to run_install()"
+            raise ValueError(msg)
 
         return self._run(
             *args,
@@ -592,19 +617,20 @@ class Session:
         if callable(args[0]):
             return self._run_func(args[0], args[1:])  # type: ignore[unreachable]
 
-        # Using `"uv"` when `uv` is the backend is guaranteed to work, even if it was co-installed with nox.
-        if (
-            self.virtualenv.venv_backend == "uv"
-            and args[0] == "uv"
-            and nox.virtualenv.UV != "uv"
-            and shutil.which("uv", path=self.bin) is None  # Session uv takes priority
-        ):
-            args = (nox.virtualenv.UV, *args[1:])
+        # Using `"uv"` or `"uvx" when `uv` is the backend is guaranteed to
+        # work, even if it was co-installed with nox.
+        if self.virtualenv.venv_backend == "uv" and nox.virtualenv.UV != "uv":
+            if (
+                args[0] == "uv"
+                and shutil.which("uv", path=self.bin)
+                is None  # Session uv takes priority
+            ):
+                args = (nox.virtualenv.UV, *args[1:])
+            elif args[0] == "uvx" and shutil.which("uvx", path=self.bin) is None:
+                args = (f"{nox.virtualenv.UV}x", *args[1:])
 
         # Combine the env argument with our virtualenv's env vars.
-        if include_outer_env:
-            overlay_env = env or {}
-            env = {**self.env, **overlay_env}
+        env = self.virtualenv._get_env(env or {}, include_outer_env=include_outer_env)
 
         # If --error-on-external-run is specified, error on external programs.
         if self._runner.global_config.error_on_external_run and external is None:
@@ -691,13 +717,15 @@ class Session:
         if isinstance(venv, CondaEnv):
             prefix_args = ("--prefix", venv.location)
         elif not isinstance(venv, PassthroughEnv):
-            raise ValueError(
+            msg = (
                 "A session without a conda environment can not install dependencies"
                 " from conda."
             )
+            raise TypeError(msg)
 
         if not args:
-            raise ValueError("At least one argument required to install().")
+            msg = "At least one argument required to install()."
+            raise ValueError(msg)
 
         if self._runner.global_config.no_install and (
             isinstance(venv, PassthroughEnv) or venv._reused
@@ -705,10 +733,11 @@ class Session:
             return
 
         # Escape args that should be (conda-specific; pip install does not need this)
-        args = _dblquote_pkg_install_args(args)
+        if sys.platform.startswith("win32"):
+            args = _dblquote_pkg_install_args(args)
 
         if silent is None:
-            silent = True
+            silent = not self._runner.global_config.verbose
 
         extraopts: list[str] = []
         if auto_offline and venv.is_offline():
@@ -746,12 +775,12 @@ class Session:
     def install(
         self,
         *args: str,
-        env: Mapping[str, str] | None = None,
+        env: Mapping[str, str | None] | None = None,
         include_outer_env: bool = True,
         silent: bool | None = None,
         success_codes: Iterable[int] | None = None,
         log: bool = True,
-        external: ExternalType | None = None,
+        external: ExternalType | None = None,  # noqa: ARG002
         stdout: int | IO[str] | None = None,
         stderr: int | IO[str] | None = subprocess.STDOUT,
         interrupt_timeout: float | None = DEFAULT_INTERRUPT_TIMEOUT,
@@ -798,26 +827,27 @@ class Session:
         if not isinstance(
             venv, (CondaEnv, VirtualEnv, PassthroughEnv)
         ):  # pragma: no cover
-            raise ValueError(
-                "A session without a virtualenv can not install dependencies."
-            )
+            msg = f"A session without a virtualenv (got {venv!r}) can not install dependencies."
+            raise TypeError(msg)
         if isinstance(venv, PassthroughEnv):
             if self._runner.global_config.no_install:
                 return
-            raise ValueError(
+            msg = (
                 f"Session {self.name} does not have a virtual environment, so use of"
                 " session.install() is no longer allowed since it would modify the"
                 " global Python environment. If you're really sure that is what you"
                 ' want to do, use session.run("pip", "install", ...) instead.'
             )
+            raise ValueError(msg)
         if not args:
-            raise ValueError("At least one argument required to install().")
+            msg = "At least one argument required to install()."
+            raise ValueError(msg)
 
         if self._runner.global_config.no_install and venv._reused:
             return
 
         if silent is None:
-            silent = True
+            silent = not self._runner.global_config.verbose
 
         if isinstance(venv, VirtualEnv) and venv.venv_backend == "uv":
             cmd = ["uv", "pip", "install"]
@@ -904,6 +934,8 @@ class SessionRunner:
         func: Func,
         global_config: argparse.Namespace,
         manifest: Manifest,
+        *,
+        multi: bool = False,
     ) -> None:
         self.name = name
         self.signatures = signatures
@@ -912,13 +944,22 @@ class SessionRunner:
         self.manifest = manifest
         self.venv: ProcessEnv | None = None
         self.posargs: list[str] = global_config.posargs[:]
+        self.result: Result | None = None
+        self.multi = multi
+
+        if getattr(func, "parametrize", None):
+            self.multi = True
+
+    def __repr__(self) -> str:
+        return (
+            f"<{self.__class__.__name__} {self.name}: {self.signatures!r} {self.multi}>"
+        )
 
     @property
     def description(self) -> str | None:
         doc = self.func.__doc__
         if doc:
-            first_line = doc.strip().split("\n")[0]
-            return first_line
+            return doc.strip().split("\n")[0]
         return None
 
     def __str__(self) -> str:
@@ -937,6 +978,44 @@ class SessionRunner:
     def envdir(self) -> str:
         return _normalize_path(self.global_config.envdir, self.friendly_name)
 
+    def get_direct_dependencies(
+        self, sessions_by_id: Mapping[str, SessionRunner] | None = None
+    ) -> Iterator[SessionRunner]:
+        """Yields the sessions of the session's direct dependencies.
+
+        Args:
+            sessions_by_id (Mapping[str, ~nox.sessions.SessionRunner] | None): An
+                optional mapping from both dependency signatures and names to
+                corresponding ``SessionRunner``s. If this is not provided,
+                ``self.manifest.all_sessions_by_signature`` will be used to find the
+                sessions corresponding to signatures in ``self.func.requires``, and
+                non-signature names (i.e. names of sessions that were parameterized with
+                multiple Pythons) in ``self.func.requires`` will be resolved via
+                ``self.manifest.parametrized_sessions_by_name``.
+
+        Returns:
+            Iterator[~nox.session.SessionRunner]
+
+        Raises:
+            KeyError: If a dependency's session could not be found.
+        """
+        try:
+            if sessions_by_id is None:
+                sessions_by_signature = self.manifest.all_sessions_by_signature
+                parametrized_sessions_by_name = (
+                    self.manifest.parametrized_sessions_by_name
+                )
+                for requirement in self.func.requires:
+                    if requirement in sessions_by_signature:
+                        yield sessions_by_signature[requirement]
+                    else:
+                        yield from parametrized_sessions_by_name[requirement]
+            else:
+                yield from map(sessions_by_id.__getitem__, self.func.requires)
+        except KeyError as exc:
+            msg = f"Session not found: {exc.args[0]}"
+            raise KeyError(msg) from exc
+
     def _create_venv(self) -> None:
         reuse_existing = self.reuse_existing_venv()
 
@@ -947,34 +1026,18 @@ class SessionRunner:
             or "virtualenv"
         ).split("|")
 
-        # Support fallback backends
-        for bk in backends:
-            if bk not in nox.virtualenv.ALL_VENVS:
-                msg = f"Expected venv_backend one of {list(nox.virtualenv.ALL_VENVS)!r}, but got {bk!r}."
-                raise ValueError(msg)
+        download_python = (
+            self.global_config.download_python or self.func.download_python or "auto"
+        )
 
-        for bk in backends[:-1]:
-            if bk not in nox.virtualenv.OPTIONAL_VENVS:
-                msg = f"Only optional backends ({list(nox.virtualenv.OPTIONAL_VENVS)!r}) may have a fallback, {bk!r} is not optional."
-                raise ValueError(msg)
-
-        for bk in backends:
-            if nox.virtualenv.OPTIONAL_VENVS.get(bk, True):
-                backend = bk
-                break
-        else:
-            msg = f"No backends present, looked for {backends!r}."
-            raise ValueError(msg)
-
-        if backend == "none" or self.func.python is False:
-            self.venv = nox.virtualenv.ALL_VENVS["none"]()
-        else:
-            self.venv = nox.virtualenv.ALL_VENVS[backend](
-                self.envdir,
-                interpreter=self.func.python,
-                reuse_existing=reuse_existing,
-                venv_params=self.func.venv_params,
-            )
+        self.venv = get_virtualenv(
+            *backends,
+            download_python=download_python,
+            envdir=self.envdir,
+            reuse_existing=reuse_existing,
+            interpreter=self.func.python,
+            venv_params=self.func.venv_params,
+        )
 
         self.venv.create()
 
@@ -1009,6 +1072,12 @@ class SessionRunner:
         Returns:
             bool: True if the existing virtual environment should be reused, False otherwise.
         """
+        if self.global_config.reuse_venv not in {"always", "never", "no", "yes", None}:
+            msg = (
+                "nox.options.reuse_venv must be set to 'always', 'never', 'no', or 'yes',"
+                f" got {self.global_config.reuse_venv!r}!"
+            )
+            raise AttributeError(msg)
 
         return any(
             (
@@ -1024,8 +1093,22 @@ class SessionRunner:
         )
 
     def execute(self) -> Result:
-        logger.warning(f"Running session {self.friendly_name}")
+        logger.session_info(f"Running session {self.friendly_name}")
 
+        for dependency in self.get_direct_dependencies():
+            if not dependency.result:
+                self.result = Result(
+                    self,
+                    Status.ABORTED,
+                    reason=(
+                        f"Prerequisite session {dependency.friendly_name} was not"
+                        " successful"
+                    ),
+                    duration=0,
+                )
+                return self.result
+
+        start = time.perf_counter()
         try:
             cwd = os.path.realpath(os.path.dirname(self.global_config.noxfile))
 
@@ -1036,25 +1119,49 @@ class SessionRunner:
                 self.func(session)
 
             # Nothing went wrong; return a success.
-            return Result(self, Status.SUCCESS)
+            self.result = Result(
+                self, Status.SUCCESS, duration=time.perf_counter() - start
+            )
 
         except nox.virtualenv.InterpreterNotFound as exc:
             if self.global_config.error_on_missing_interpreters:
-                return Result(self, Status.FAILED, reason=str(exc))
+                self.result = Result(
+                    self,
+                    Status.FAILED,
+                    reason=str(exc),
+                    duration=time.perf_counter() - start,
+                )
             else:
                 logger.warning(
                     "Missing interpreters will error by default on CI systems."
                 )
-                return Result(self, Status.SKIPPED, reason=str(exc))
+                self.result = Result(
+                    self,
+                    Status.SKIPPED,
+                    reason=str(exc),
+                    duration=time.perf_counter() - start,
+                )
 
         except _SessionQuit as exc:
-            return Result(self, Status.ABORTED, reason=str(exc))
+            self.result = Result(
+                self,
+                Status.ABORTED,
+                reason=str(exc),
+                duration=time.perf_counter() - start,
+            )
 
         except _SessionSkip as exc:
-            return Result(self, Status.SKIPPED, reason=str(exc))
+            self.result = Result(
+                self,
+                Status.SKIPPED,
+                reason=str(exc),
+                duration=time.perf_counter() - start,
+            )
 
         except nox.command.CommandFailed:
-            return Result(self, Status.FAILED)
+            self.result = Result(
+                self, Status.FAILED, duration=time.perf_counter() - start
+            )
 
         except KeyboardInterrupt:
             logger.error(f"Session {self.friendly_name} interrupted.")
@@ -1062,14 +1169,33 @@ class SessionRunner:
 
         except Exception as exc:
             logger.exception(f"Session {self.friendly_name} raised exception {exc!r}")
-            return Result(self, Status.FAILED)
+            self.result = Result(
+                self, Status.FAILED, duration=time.perf_counter() - start
+            )
+
+        return self.result
+
+
+def _duration_str(seconds: float, text: str) -> str:
+    time_str = humanize.naturaldelta(datetime.timedelta(seconds=seconds))
+
+    # Might be "a moment" if short, return empty string in that case
+    if time_str == "a moment":
+        return ""
+
+    return text.format(time=time_str)
 
 
 class Result:
     """An object representing the result of a session."""
 
     def __init__(
-        self, session: SessionRunner, status: Status, reason: str | None = None
+        self,
+        session: SessionRunner,
+        status: Status,
+        reason: str | None = None,
+        *,
+        duration: float = 0.0,
     ) -> None:
         """Initialize the Result object.
 
@@ -1078,17 +1204,16 @@ class Result:
                 The session runner which ran.
             status (~nox.sessions.Status): The final result status.
             reason (str): Additional info.
+            duration (float): Time taken in seconds.
         """
         self.session = session
         self.status = status
         self.reason = reason
+        self.duration = duration
 
     def __bool__(self) -> bool:
         return self.status.value > 0
 
-    def __nonzero__(self) -> bool:
-        return self.__bool__()
-
     @property
     def imperfect(self) -> str:
         """Return the English imperfect tense for the status.
@@ -1097,12 +1222,12 @@ class Result:
             str: A word or phrase representing the status.
         """
         if self.status == Status.SUCCESS:
-            return "was successful"
+            return "was successful" + _duration_str(self.duration, " in {time}")
 
         status = self.status.name.lower()
         if self.reason:
-            return f"{status}: {self.reason}"
-
+            duration_err = _duration_str(self.duration, " (took {time})")
+            return f"{status}: {self.reason}{duration_err}"
         return status
 
     def log(self, message: str) -> None:
@@ -1132,4 +1257,5 @@ class Result:
             "result": self.status.name.lower(),
             "result_code": self.status.value,
             "signatures": self.session.signatures,
+            "duration": self.duration,
         }
