@@ -19,15 +19,14 @@ import contextlib
 import functools
 import json
 import os
-import platform
 import re
 import shutil
 import subprocess
 import sys
-from collections.abc import Callable, Mapping
+import sysconfig
 from pathlib import Path
 from socket import gethostbyname
-from typing import Any, ClassVar
+from typing import TYPE_CHECKING, Any, ClassVar, Literal
 
 from packaging import version
 
@@ -35,6 +34,39 @@ import nox
 import nox.command
 from nox.logger import logger
 
+if TYPE_CHECKING:
+    from collections.abc import Callable, Mapping, Sequence
+
+    from nox._typing import Python
+
+__all__ = [
+    "ALL_VENVS",
+    "HAS_UV",
+    "OPTIONAL_VENVS",
+    "UV",
+    "UV_VERSION",
+    "CondaEnv",
+    "InterpreterNotFound",
+    "PassthroughEnv",
+    "ProcessEnv",
+    "VirtualEnv",
+    "find_uv",
+    "get_virtualenv",
+    "pbs_install_python",
+    "uv_install_python",
+    "uv_version",
+]
+
+
+def __dir__() -> list[str]:
+    return __all__
+
+
+# Use for test mocking and to make mypy happy
+_PLATFORM = sys.platform
+_IS_MINGW = sysconfig.get_platform().startswith("mingw")
+
+
 # Problematic environment variables that are stripped from all commands inside
 # of a virtualenv. See https://github.com/theacodes/nox/issues/44
 _BLACKLISTED_ENV_VARS = frozenset(
@@ -43,49 +75,96 @@ _BLACKLISTED_ENV_VARS = frozenset(
         "PIP_REQUIRE_VIRTUALENV",
         "__PYVENV_LAUNCHER__",
         "UV_SYSTEM_PYTHON",
+        "UV_PYTHON",
     ]
 )
-_SYSTEM = platform.system()
+
+NOX_PBS_PYTHONS = Path.home() / ".local" / "share" / "nox" / "pythons"
 
 
-def find_uv() -> tuple[bool, str]:
-    uv_on_path = shutil.which("uv")
+def find_uv() -> tuple[bool, str, version.Version]:
+    uv_name = os.environ.get("UV", None)
+    uv_on_path = shutil.which(uv_name or "uv")
 
     # Look for uv in Nox's environment, to handle `pipx install nox[uv]`.
-    with contextlib.suppress(ImportError, FileNotFoundError):
-        from uv import find_uv_bin
+    if uv_name is None:
+        with contextlib.suppress(ImportError, FileNotFoundError):
+            from uv import find_uv_bin  # noqa: PLC0415
 
-        uv_bin = find_uv_bin()
+            uv_bin = find_uv_bin()
 
-        # If the returned value is the same as calling "uv" already, don't
-        # expand (simpler logging)
-        if uv_on_path and Path(uv_bin).samefile(uv_on_path):
-            return True, "uv"
+            uv_vers = uv_version(uv_bin)
+            if uv_vers > version.Version("0"):
+                # If the returned value is the same as calling "uv" already, don't
+                # expand (simpler logging)
+                if uv_on_path and Path(uv_bin).samefile(uv_on_path):
+                    return True, "uv", uv_vers
 
-        return True, uv_bin
+                return True, uv_bin, uv_vers
 
     # Fall back to PATH.
-    return uv_on_path is not None, "uv"
+    uv_vers = uv_version(uv_name or "uv")
+    return (
+        uv_on_path is not None and uv_vers > version.Version("0"),
+        uv_name or "uv",
+        uv_vers,
+    )
 
 
-def uv_version() -> version.Version:
+def _find_python(interpreter: str, xy_ver: str) -> str | None:
+    """Find a python executable matching the requested interpreter"""
+    if shutil.which(interpreter):
+        return interpreter
+
+    # Windows only search for the executable
+    if _PLATFORM.startswith("win"):
+        # Allow versions of the form ``X.Y-32`` for Windows.
+        match = re.match(r"^\d\.\d+-32?$", interpreter)
+        if match:
+            # preserve the "-32" suffix, as the Python launcher expects it.
+            xy_ver = interpreter
+
+        path_from_launcher = locate_via_py(xy_ver)
+        if path_from_launcher:
+            return path_from_launcher
+
+        path_from_version_param = locate_using_path_and_version(xy_ver)
+        if path_from_version_param:
+            return path_from_version_param
+
+    # not found case
+    return None
+
+
+def uv_version(uv_bin: str) -> version.Version:
     """Returns uv's version defaulting to 0.0 if uv is not available"""
     try:
         ret = subprocess.run(
-            [UV, "version", "--output-format", "json"],
+            [uv_bin, "self", "version", "--output-format", "json"],
             check=False,
             text=True,
             capture_output=True,
+            encoding="utf-8",
         )
-    except FileNotFoundError:
+    except (FileNotFoundError, PermissionError):
         logger.info("uv binary not found.")
         return version.Version("0.0")
 
+    if ret.returncode == 2:
+        # uv < 0.7
+        ret = subprocess.run(
+            [uv_bin, "version", "--output-format", "json"],
+            check=False,
+            text=True,
+            capture_output=True,
+            encoding="utf-8",
+        )
+
     if ret.returncode == 0 and ret.stdout:
         return version.Version(json.loads(ret.stdout).get("version"))
-    else:
-        logger.info("Failed to establish uv's version.")
-        return version.Version("0.0")
+
+    logger.info("Failed to establish uv's version.")
+    return version.Version("0.0")
 
 
 def uv_install_python(python_version: str) -> bool:
@@ -97,10 +176,70 @@ def uv_install_python(python_version: str) -> bool:
     return ret.returncode == 0
 
 
-HAS_UV, UV = find_uv()
-# supported since uv 0.3 but 0.4.16 is the first version that doesn't cause
-# issues for nox with pypy/cpython confusion
-UV_PYTHON_SUPPORT = uv_version() >= version.Version("0.4.16")
+def _find_pbs_python(implementation: str, version: str) -> str | None:
+    """Check for an existing pbs-installer installation
+    by default it creates dirs with this format:
+    "pypy@3.8.16", "cpython@3.13.3" """
+    executable = "python.exe" if _PLATFORM.startswith("win") else "bin/python"
+
+    if NOX_PBS_PYTHONS.exists():
+        for path in NOX_PBS_PYTHONS.iterdir():
+            if path.is_dir() and path.name.startswith(f"{implementation}@{version}."):
+                python_exe = path / executable
+                if python_exe.exists():
+                    return str(python_exe)
+    return None
+
+
+def pbs_install_python(python_version: str) -> str | None:
+    """Attempts to install a given python version with pbs-installer.
+
+    Returns the full path to the installed executable, or None if installation failed.
+    """
+
+    # separate implementation / xyz version
+    match = re.match(
+        r"^(?P<impl>pypy|cpython|python)?[-_\.]?(?P<xyz_ver>\d(\.\d+)?(\.\d+))?$",
+        python_version,
+        re.IGNORECASE,
+    )
+
+    if not match:
+        logger.warning(f"{python_version=} is not a valid version to install with pbs")
+        return None
+
+    implementation: Literal["cpython", "pypy"] = (
+        "cpython" if match.group("impl").lower() in ("cpython", "python") else "pypy"
+    )
+    xyz_ver = match.group("xyz_ver")
+
+    if python_exe := _find_pbs_python(implementation, xyz_ver):
+        return python_exe
+
+    # Requires the [pbs] extra to make it past here
+    try:
+        import pbs_installer  # noqa: PLC0415
+    except ModuleNotFoundError:  # pragma: nocover
+        logger.warning(
+            "Nox was installed without the `[pbs]` extra, can't download Python"
+        )
+        return None
+
+    try:
+        pbs_installer.install(
+            xyz_ver,
+            destination=NOX_PBS_PYTHONS,
+            version_dir=True,
+            implementation=implementation,
+        )
+    except Exception as err:
+        logger.warning(f"Failed to install a pbs version for {python_version=}: {err}")
+        return None
+
+    return _find_pbs_python(implementation, xyz_ver)
+
+
+HAS_UV, UV, UV_VERSION = find_uv()
 
 
 class InterpreterNotFound(OSError):
@@ -121,24 +260,15 @@ class ProcessEnv(abc.ABC):
     allowed_globals: ClassVar[tuple[Any, ...]] = ()
 
     def __init__(
-        self, bin_paths: None = None, env: Mapping[str, str | None] | None = None
+        self,
+        bin_paths: Sequence[str] | None = None,
+        env: Mapping[str, str | None] | None = None,
     ) -> None:
-        self._bin_paths = bin_paths
+        self._bin_paths = None if bin_paths is None else list(bin_paths)
         self._reused = False
 
-        # Filter envs now so `.env` is dict[str, str] (easier to use)
-        # even though .command's env supports None.
-        env = env or {}
-        env = {**os.environ, **env}
-        self.env = {k: v for k, v in env.items() if v is not None}
-
-        for key in _BLACKLISTED_ENV_VARS:
-            self.env.pop(key, None)
-
-        if self.bin_paths:
-            self.env["PATH"] = os.pathsep.join(
-                [*self.bin_paths, self.env.get("PATH", "")]
-            )
+        # .command's env supports None, meaning don't include value even if in parent
+        self.env = {**dict.fromkeys(_BLACKLISTED_ENV_VARS), **(env or {})}
 
     @property
     def bin_paths(self) -> list[str] | None:
@@ -149,7 +279,8 @@ class ProcessEnv(abc.ABC):
         """The first bin directory for the virtualenv."""
         paths = self.bin_paths
         if paths is None:
-            raise ValueError("The environment does not have a bin directory.")
+            msg = "The environment does not have a bin directory."
+            raise ValueError(msg)
         return paths[0]
 
     @abc.abstractmethod
@@ -166,6 +297,27 @@ class ProcessEnv(abc.ABC):
         Returns the string used to select this environment.
         """
 
+    def _get_env(
+        self,
+        /,
+        env: Mapping[str, str | None],
+        *,
+        include_outer_env: bool = True,
+    ) -> dict[str, str | None]:
+        """
+        Get the computed environment, with bin paths added.  You can request
+        the outer environment be excluded. The initial env can be empty.
+        """
+
+        computed_env = {**self.env, **env}
+        if include_outer_env:
+            computed_env = {**os.environ, **computed_env}
+        if self.bin_paths:
+            computed_env["PATH"] = os.pathsep.join(
+                [*self.bin_paths, computed_env.get("PATH") or ""]
+            )
+        return computed_env
+
 
 def locate_via_py(version: str) -> str | None:
     """Find the Python executable using the Windows Launcher.
@@ -193,6 +345,7 @@ def locate_via_py(version: str) -> str | None:
             check=False,
             text=True,
             capture_output=True,
+            encoding="utf-8",
         )
         if ret.returncode == 0 and ret.stdout:
             return ret.stdout.strip()
@@ -286,10 +439,11 @@ class CondaEnv(ProcessEnv):
         self,
         location: str,
         interpreter: str | None = None,
-        reuse_existing: bool = False,
-        venv_params: Any = None,
         *,
+        reuse_existing: bool = False,
+        venv_params: Sequence[str] = (),
         conda_cmd: str = "conda",
+        **kwargs: Any,
     ):
         self.location_name = location
         self.location = os.path.abspath(location)
@@ -306,7 +460,7 @@ class CondaEnv(ProcessEnv):
             if self.reuse_existing and is_conda:
                 return False
             if not is_conda:
-                shutil.rmtree(self.location)
+                shutil.rmtree(self.location, ignore_errors=True)
             else:
                 cmd = [
                     self.conda_cmd,
@@ -318,8 +472,7 @@ class CondaEnv(ProcessEnv):
                 ]
                 nox.command.run(cmd, silent=True, log=False)
             # Make sure that location is clean
-            with contextlib.suppress(FileNotFoundError):
-                shutil.rmtree(self.location)
+            shutil.rmtree(self.location, ignore_errors=True)
 
         return True
 
@@ -327,7 +480,7 @@ class CondaEnv(ProcessEnv):
     def bin_paths(self) -> list[str]:
         """Returns the location of the conda env's bin folder."""
         # see https://github.com/conda/conda/blob/f60f0f1643af04ed9a51da3dd4fa242de81e32f4/conda/activate.py#L563-L572
-        if _SYSTEM == "Windows":
+        if _PLATFORM.startswith("win"):
             return [
                 self.location,
                 os.path.join(self.location, "Library", "mingw-w64", "bin"),
@@ -342,7 +495,7 @@ class CondaEnv(ProcessEnv):
     def create(self) -> bool:
         """Create the conda env."""
         if not self._clean_location():
-            logger.debug(f"Re-using existing conda env at {self.location_name}.")
+            logger.debug(f"Reusing existing conda env at {self.location_name}.")
 
             self._reused = True
 
@@ -384,9 +537,9 @@ class CondaEnv(ProcessEnv):
         try:
             # DNS resolution to detect situation (1) or (2).
             host = gethostbyname("repo.anaconda.com")
-            return host is None
         except BaseException:  # pragma: no cover
             return True
+        return host is None
 
     @property
     def venv_backend(self) -> str:
@@ -406,6 +559,7 @@ class VirtualEnv(ProcessEnv):
               be ``py -3.6-32``
             * ``X.Y.Z``, e.g. ``3.4.9``
             * ``pythonX.Y``, e.g. ``python2.7``
+            * ``pypyX.Y``, e.g. ``pypy3.10`` (also ``pypy-3.10`` allowed)
             * A path in the filesystem to a Python executable
 
             If not specified, this will use the currently running Python.
@@ -414,17 +568,22 @@ class VirtualEnv(ProcessEnv):
     """
 
     is_sandboxed = True
-    allowed_globals = (UV,)
+    allowed_globals = (UV, f"{UV}x")
 
     def __init__(
         self,
         location: str,
         interpreter: str | None = None,
-        reuse_existing: bool = False,
         *,
+        download_python: Literal["auto", "never", "always"] = "auto",
+        reuse_existing: bool = False,
         venv_backend: str = "virtualenv",
-        venv_params: Any = None,
+        venv_params: Sequence[str] = (),
     ):
+        # "pypy-" -> "pypy"
+        if interpreter and interpreter.startswith("pypy-"):
+            interpreter = interpreter[:4] + interpreter[5:]
+
         self.location_name = location
         self.location = os.path.abspath(location)
         self.interpreter = interpreter
@@ -432,8 +591,9 @@ class VirtualEnv(ProcessEnv):
         self.reuse_existing = reuse_existing
         self._venv_backend = venv_backend
         self.venv_params = venv_params or []
+        self.download_python = download_python
         if venv_backend not in {"virtualenv", "venv", "uv"}:
-            msg = f"venv_backend {venv_backend} not recognized"
+            msg = f"venv_backend {venv_backend!r} not recognized"
             raise ValueError(msg)
         super().__init__(env={"VIRTUAL_ENV": self.location, "CONDA_PREFIX": None})
 
@@ -446,13 +606,15 @@ class VirtualEnv(ProcessEnv):
                 and self._check_reused_environment_interpreter()
             ):
                 return False
-            shutil.rmtree(self.location)
+            # uv clears it for us, and it balks at files left around
+            if self.venv_backend != "uv":
+                shutil.rmtree(self.location, ignore_errors=True)
         return True
 
     def _read_pyvenv_cfg(self) -> dict[str, str] | None:
         """Read a pyvenv.cfg file into dict, returns None if missing."""
         path = os.path.join(self.location, "pyvenv.cfg")
-        with contextlib.suppress(FileNotFoundError), open(path) as fp:
+        with contextlib.suppress(FileNotFoundError), open(path, encoding="utf-8") as fp:
             parts = (x.partition("=") for x in fp if "=" in x)
             return {k.strip(): v.strip() for k, _, v in parts}
         return None
@@ -551,54 +713,63 @@ class VirtualEnv(ProcessEnv):
 
         # If this is just a X, X.Y, or X.Y.Z string, extract just the X / X.Y
         # part and add Python to the front of it.
-        match = re.match(r"^(?P<xy_ver>\d(\.\d+)?)(\.\d+)?$", self.interpreter)
+        match = re.match(r"^(?P<xy_ver>\d(\.\d+)?)(\.\d+)?(?P<t>t?)$", self.interpreter)
         if match:
             xy_version = match.group("xy_ver")
-            cleaned_interpreter = f"python{xy_version}"
+            t = match.group("t")
+            cleaned_interpreter = f"python{xy_version}{t}"
 
-        # If the cleaned interpreter is on the PATH, go ahead and return it.
-        if shutil.which(cleaned_interpreter):
-            self._resolved = cleaned_interpreter
-            return self._resolved
-
-        if HAS_UV and UV_PYTHON_SUPPORT:
-            uv_python_success = uv_install_python(cleaned_interpreter)
-            if uv_python_success:
-                self._resolved = cleaned_interpreter
+        # never -> check for interpreters
+        if self.download_python == "never":
+            if resolved := _find_python(cleaned_interpreter, xy_version):
+                self._resolved = resolved
                 return self._resolved
 
-        # The rest of this is only applicable to Windows, so if we don't have
-        # an interpreter by now, raise.
-        if _SYSTEM != "Windows":
-            self._resolved = InterpreterNotFound(self.interpreter)
-            raise self._resolved
+        # always -> skip check, always install
+        elif self.download_python == "always" and self.venv_backend == "uv":
+            if HAS_UV and version.Version("0.4.16") <= UV_VERSION:
+                uv_python_success = uv_install_python(cleaned_interpreter)
+                if uv_python_success:
+                    self._resolved = cleaned_interpreter
+                    return self._resolved
 
-        # Allow versions of the form ``X.Y-32`` for Windows.
-        match = re.match(r"^\d\.\d+-32?$", cleaned_interpreter)
-        if match:
-            # preserve the "-32" suffix, as the Python launcher expects
-            # it.
-            xy_version = cleaned_interpreter
+        elif self.download_python == "always" and self.venv_backend in (
+            "venv",
+            "virtualenv",
+        ):
+            pbs_python_path = pbs_install_python(cleaned_interpreter)
+            if pbs_python_path:
+                self._resolved = pbs_python_path
+                return self._resolved
 
-        path_from_launcher = locate_via_py(xy_version)
-        if path_from_launcher:
-            self._resolved = path_from_launcher
-            return self._resolved
+        # auto -> check interpreters -> fallback to installing
+        else:
+            if resolved := _find_python(cleaned_interpreter, xy_version):
+                self._resolved = resolved
+                return self._resolved
 
-        path_from_version_param = locate_using_path_and_version(xy_version)
-        if path_from_version_param:
-            self._resolved = path_from_version_param
-            return self._resolved
+            if (
+                self.venv_backend == "uv"
+                and HAS_UV
+                and version.Version("0.4.16") <= UV_VERSION
+            ):
+                uv_python_success = uv_install_python(cleaned_interpreter)
+                if uv_python_success:
+                    self._resolved = cleaned_interpreter
+                    return self._resolved
+            elif self.venv_backend in ("venv", "virtualenv"):
+                pbs_python_path = pbs_install_python(cleaned_interpreter)
+                if pbs_python_path:
+                    self._resolved = pbs_python_path
+                    return self._resolved
 
-        # If we got this far, then we were unable to resolve the interpreter
-        # to an actual executable; raise an exception.
         self._resolved = InterpreterNotFound(self.interpreter)
         raise self._resolved
 
     @property
     def bin_paths(self) -> list[str]:
         """Returns the location of the virtualenv's bin folder."""
-        if _SYSTEM == "Windows":
+        if _PLATFORM.startswith("win") and not _IS_MINGW:
             return [os.path.join(self.location, "Scripts")]
         return [os.path.join(self.location, "bin")]
 
@@ -606,7 +777,7 @@ class VirtualEnv(ProcessEnv):
         """Create the virtualenv or venv."""
         if not self._clean_location():
             logger.debug(
-                f"Re-using existing virtual environment at {self.location_name}."
+                f"Reusing existing virtual environment at {self.location_name}."
             )
 
             self._reused = True
@@ -614,7 +785,13 @@ class VirtualEnv(ProcessEnv):
             return False
 
         if self.venv_backend == "virtualenv":
-            cmd = [sys.executable, "-m", "virtualenv", self.location]
+            cmd = [
+                sys.executable,
+                "-m",
+                "virtualenv",
+                self.location,
+                "--no-periodic-update",
+            ]
             if self.interpreter:
                 cmd.extend(["-p", self._resolved_interpreter])
         elif self.venv_backend == "uv":
@@ -663,3 +840,42 @@ OPTIONAL_VENVS = {
     "micromamba": shutil.which("micromamba") is not None,
     "uv": HAS_UV,
 }
+
+
+def get_virtualenv(
+    *backends: str,
+    download_python: Literal["auto", "never", "always"],
+    envdir: str,
+    reuse_existing: bool,
+    interpreter: Python = None,
+    venv_params: Sequence[str] = (),
+) -> ProcessEnv:
+    # Support fallback backends
+    for bk in backends:
+        if bk not in ALL_VENVS:
+            msg = f"Expected venv_backend one of {sorted(ALL_VENVS)!r}, but got {bk!r}."
+            raise ValueError(msg)
+
+    for bk in backends[:-1]:
+        if bk not in OPTIONAL_VENVS:
+            msg = f"Only optional backends ({sorted(OPTIONAL_VENVS)!r}) may have a fallback, {bk!r} is not optional."
+            raise ValueError(msg)
+
+    for bk in backends:
+        if OPTIONAL_VENVS.get(bk, True):
+            backend = bk
+            break
+    else:
+        msg = f"No backends present, looked for {backends!r}."
+        raise ValueError(msg)
+
+    if backend == "none" or interpreter is False:
+        return ALL_VENVS["none"]()
+
+    return ALL_VENVS[backend](
+        envdir,
+        download_python=download_python,
+        interpreter=interpreter,
+        reuse_existing=reuse_existing,
+        venv_params=venv_params,
+    )
